declare function join(array: string[], fn: string[] | ((index: number) => string)): string;
declare function delay(time?: number): Promise<unknown>;
declare function getFuncArgs(fn: Function): string;
declare function isStringTemplateArray(data: any): boolean;
declare function splitTwoPart(str: string, sep: string): string[];

declare const subscribe: unique symbol;
declare const forceUpdate: unique symbol;
declare const index: unique symbol;
declare const value: unique symbol;
declare const reactValue: unique symbol;
declare const getListeners: unique symbol;
declare const json: unique symbol;
declare const replaceValue: unique symbol;

var version = "0.0.11";

/*
 * @Author: tackchen
 * @Date: 2022-10-11 16:41:40
 * @Description: Coding something
 */
interface IJson<T = any> {
    [prop: string | symbol]: T;
}

interface IBuilderParameter {
    exe: Function;
    type: 'style' | 'pseudo' |
        'text' | 'selector' | 'children' |
        'comp' | 'react' | 'builder' |
        // 控制器 没有for
        'for' | 'if' | 'show' | 'model' | 'switch' |
        'on' |
        // 组件
        'event' | 'prop' | 'slot' | 'life' | 'html' | 'text';
}

/*
 * @Author: tackchen
 * @Date: 2022-10-23 19:26:19
 * @Description: Coding something
 */


// export type TSimpleValue = number|string|boolean|null|undefined;

interface IReactBase<T = any> {
  [index]?: IReactItem<number>;
  [forceUpdate](old?: any, index?: number): void;
  [subscribe](fn: (v:T, old:T, index: number) => void):  T;
  [reactValue]: boolean;
  [getListeners](): Function[];
  [json] (): T;
}

interface IReactItem<T = any> extends IReactBase<T>{
  value: T;
  isUndefined(): boolean;
  toJSON: ()=> T | undefined; // ! 重写value的toJSON方法
  [replaceValue](v: T): void;
  type: 'reaction'
}

type TComputedFunc<T = any> = (...args: any[]) => T;


type TReactionItem<T=any> = IReactItem<T> | TComputedFunc<T> | IComputedItem<T>;

type TReactionValue<T> = T | IReactBuilder | TReactionItem<T>;

interface IComputedItem<T = any> extends IReactItem<T> {}


// export type TBaseTypes = number | boolean | string | null | undefined;
// type TReactTypes = TBaseTypes | IJson<TReactTypes> | TReactTypes[];

interface IReactObject<T = any> extends IReactBase<T> {
  // ! 本来应该使用 T，
  // ! 由于封装了一层，所以赋值时ts类型系统会报错，故使用IJson
  // ! 牺牲了 value和json返回值的类型
  get [value](): IJson;
  set [value](v: IJson);
}

/*
  ! 针对一个可能是ts的很奇怪的bug
  const list = $([{content: '1', done: false}]);
  done 会被 推断了 IReactItem<true> | IReactItem<false>
  而不是 IReactItem<boolean> 从而对其赋值时会导致出现ts报错
  ! 不能将类型“boolean”分配给类型“IReactItem<false> | IReactItem<true>”。
*/
type TParseBoolean<T> = T extends object|string|undefined|symbol|number|Function ? T: boolean;

type IReactWrap<T> = T extends object ? (IReactJson<T> & IReactObject<T>) : IReactItem<TParseBoolean<T>>;

type IReactJson<T> = {
  [prop in (keyof T)]: IReactWrap<T[prop]>;
} & IJson; // ! & IJson 为了绑定的时候不报类型错误


type TOnChangeFunc = (content: string, oldContent: string) => void;

interface IReactBindingTemplate<T=any> {
  template: string[], // TemplateStringsArray
  reactions: (TReactionItem<T>)[], // | any[], // ? 为了绑定的时候不报类型错误
}
interface IReactBindingTemplateFactory<T=any> {
  add(d: IReactBindingTemplate<T>): void;
  get(): IReactBindingTemplate<T>;
  reactive(
    onchange: TOnChangeFunc,
    needOldContent?: boolean,
  ): void;
  computed(): IComputedItem<T>;
}

type TReactContextType = 'dom-info' | 'style' | 'computed' | 'css';
// react上下文环境
interface IReactContext {
  type: TReactContextType,
} // todo

interface IReactBinding extends IReactBindingTemplate {
  context: IReactContext; // todo
}
interface IReactBuilder extends IBuilderParameter {
  type: 'react';
  exe(context: IReactContext): IReactBinding;
  modTemplate(mod: (v: string) => string, i?: number): void;
  isEmpty(): boolean;
  templateValue(): string;
}

declare global {

    // ! 这些声明是为了 赋值不报错

    // const a=$({b:{v:1}}); a.b = {v:2}
    // 但是会牺牲 a.b[value] 或者 a.b[json]() 的返回值类型
    // for ts declaration
    interface Array<T> extends IReactObject<T>{
    }
    interface String extends IReactItem<string> {}
    interface Number extends IReactItem<number> {}
    interface Boolean extends IReactItem<boolean> {}
    // type IHackReactObject = {
    //     [prop in any]: any;
    // }

    interface Object extends IReactObject<any>{
    }
}

/*
 * @Author: tackchen
 * @Date: 2022-10-23 19:19:06
 * @Description: Coding something
 */


type IStyleComponent = string | IStyleBuilder | IStyleAtoms | IReactBuilder;

type IStyleComponentArray = IStyleComponent | IStyleComponentArray[];

interface IStyleBuilder extends IBuilderParameter{
  exe(parent: HTMLElement): void;
  mount(parent: HTMLElement | string): void;
  // generateBindings(onchange: (onctent: string)=>void): string;
  generate(start: number): {scopeTemplate: string, scopeReactions: TReactionItem[]};
  react(): TComputedFunc<string>;
  type: 'style';
}

type TUnit = 'px' | '%' | 'rm' | 'vh' | 'vw' | 'em' | 'rem' | 'in' | 'cm' | 'mm' | 'ex' | 'pt' | 'pc';

type TI = 'i';

type TStyleValue<T = number | string> = TReactionValue<T>;

type TCssCommonValue = 'inherit' | 'initial' | 'unset' | 'revert' | 'none' | 'auto';

type TNumberStyle = (v: TStyleValue, unit?: TUnit | TI, i?: TI) => IStyleAtoms;
type TNumberAutoStyle = (v: TStyleValue|'auto', unit?: TUnit | TI, i?: TI) => IStyleAtoms;
type TPureNumberStyle = (v: TStyleValue<number>, unit?: TUnit | TI, i?: TI) => IStyleAtoms;

type TNoneArgStyle = (i?: TI) => IStyleAtoms;
type TStringStyle<T = string> = (v: T | TStyleValue<T>, i?: TI) => IStyleAtoms;
interface TColorStyle {
    (r: TReactionValue<number>, g:  TReactionValue<number>, b:  TReactionValue<number>, a?:  TReactionValue<number>, i?: TI): IStyleAtoms;
    (v:  TReactionValue<string>, i?: TI): IStyleAtoms;
}
interface TFourValueStyle {
    (v: TStyleValue, unit?: TUnit, i?: TI): IStyleAtoms;
    (top: TStyleValue, right: TStyleValue, unit?: TUnit, i?: TI): IStyleAtoms;
    (top: TStyleValue, right: TStyleValue, bottom: TStyleValue, left: TStyleValue, unit?: TUnit, i?: TI): IStyleAtoms;
}

type TTextDeco = 'blink'|'dashed'|'dotted'|'double'|'line-through'|'overline'|'solid'|'underline'|'wavy'|TCssCommonValue;
type TPosition = 'relative' | 'absolute' | 'fixed' | 'sticky' | 'static' | TCssCommonValue;

interface INoneArgsAtoms {
  // none arg style
  borderBox: TNoneArgStyle;
  relative: TNoneArgStyle;
  absolute: TNoneArgStyle;
  fixed: TNoneArgStyle;
}

interface IComposeStyle {
  cursorUrl: (...args: TReactionValue<string|number>[]) => IStyleAtoms;
}

type TStyleJsonValue = {[key in keyof IStyleArgsAtomsBase]?: TReactionValue<string|number>} & IJson<TReactionValue<string|number>>

interface IAtomsTool {
  join: (...styles: (IStyleComponent|TStyleJsonValue)[]) => IStyleAtoms;
}

interface IStyleArgsAtomsBase {
  // ([a-zA-Z]*?)(: [a-zA-Z<>]*?;\n) => '$1', 正则
  // number style
  paddingTop: TNumberStyle;
  paddingBottom: TNumberStyle;
  paddingLeft: TNumberStyle;
  paddingRight: TNumberStyle;
  marginTop: TNumberStyle;
  marginBottom: TNumberStyle;
  marginLeft: TNumberStyle;
  marginRight: TNumberStyle;
  fontSize: TNumberStyle;
  lineHeight: TNumberStyle;
  top: TNumberStyle;
  left: TNumberStyle;
  bottom: TNumberStyle;
  right: TNumberStyle;
  borderRadius: TNumberStyle;
  textIndent: TNumberStyle;

  width: TNumberAutoStyle;
  maxWidth: TNumberAutoStyle;
  height: TNumberAutoStyle;
  maxHeight: TNumberAutoStyle;
  flexBasis: TNumberAutoStyle;

  // pure number style
  opacity: TPureNumberStyle;
  zIndex: TPureNumberStyle;
  flex: TPureNumberStyle;
  flexGrow: TPureNumberStyle;
  flexShrink: TPureNumberStyle;

  // fournumber style
  margin: TFourValueStyle;
  padding: TFourValueStyle;

  // optional string style
  textDecoration: TStringStyle<TTextDeco>;
  position: TStringStyle<TPosition>;
  alignItems: TStringStyle<'stretch'|'center'|'flex-start'|'flex-end'|'baseline'|TCssCommonValue>;
  justifyContent: TStringStyle<'flex-start'|'flex-end'|'center'|'space-between'|'space-evenly'|'space-around'|TCssCommonValue>;
  display: TStringStyle<'none'|'flex'|'block'|'inline'|'inline-block'|'list-item'|'run-in'|'compact'|'marker'|'table'|'inline-table'|'table-row-group'|'table-header-group'|'table-footer-group'|'table-row'|'table-column-group'|'table-column'|'table-cell'|'table-caption'|TCssCommonValue>;
  alignContent: TStringStyle<'stretch'|'center'|'flex-start'|'flex-end'|'space-between'|'space-around'|TCssCommonValue>;
  backgroundAttachment: TStringStyle<'scroll'|'fixed'|'local'|TCssCommonValue>;
  backgroundBlendMode: TStringStyle<'normal'|'multiply'|'screen'|'overlay'|'darken'|'lighten'|'color-dodge'|'saturation'|'color'|'luminosity'|TCssCommonValue>;
  backgroundClip: TStringStyle<'border-box'|'padding-box'|'content-box'|TCssCommonValue>;
  backgroundOrigin: TStringStyle<'border-box'|'padding-box'|'content-box'|TCssCommonValue>;
  backgroundRepeat: TStringStyle<'repea'|'repeat-x'|'repeat-y'|'no-repeat'|TCssCommonValue>;
  boxSizing: TStringStyle<'content-box'|'border-box'|'inherit'>;
  clear: TStringStyle<'left'|'right'|'both'|'none'|'inherit'>;
  textAlign: TStringStyle<'auto'|'left'|'right'|'center'|'justify'|'start'|'end'|TCssCommonValue>;
  wordWrap: TStringStyle<'normal'|'break-word'>;
  whiteSpace: TStringStyle<'normal'|'pre'|'nowrap'|'pre-wrap'|'pre-line'|'inherit'>;
  wordBreak: TStringStyle<'normal'|'break-all'|'keep-all'>;
  wordSpacing: TStringStyle<'normal'|'length'|'inherit'>;
  verticalAlign: TStringStyle<'baseline'|'sub'|'super'|'top'|'text-top'|'middle'|'bottom'|'text-bottom'|'length'|'%'|'inherit'>;
  fontStyle: TStringStyle<'normal'|'italic'|'oblique'|'inherit'>;
  flexDirection: TStringStyle<'row'|'row-reverse'|'column'|'column-reverse'|'initial'|'inherit'>;
  flexWrap: TStringStyle<'nowrap'|'wrap'|'wrap-reverse'|'initial'|'inherit'>;
  resize: TStringStyle<'none'|'both'|'horizontal'|'vertical'>;
  textOverflow: TStringStyle<'clip'|'ellipsis'|'string'|'initial'|'inherit'>;
  float: TStringStyle<'left'|'right'|'none'|'inherit'>;
  visibility: TStringStyle<'visible'|'hidden'|'collapse'|'inherit'>;
  overflow: TStringStyle<'visible'|'hidden'|'scroll'|'auto'|'inherit'>;
  overflowX: TStringStyle<'visible'|'hidden'|'scroll'|'auto'|'inherit'>;
  overflowY: TStringStyle<'visible'|'hidden'|'scroll'|'auto'|'inherit'>;
  cursor: TStringStyle<'auto'|'crosshair'|'pointer'|'move'|'e-resize'|'ne-resize'|'nw-resize'|'n-resize'|'se-resize'|'sw-resize'|'s-resize'|'w-resize'|'text'|'wait'|'help'>;
  
  // ([a-z\-]+)\|
  // *([a-z\-]+)	.*\n

  // common string style
  border: TStringStyle;
  borderBottom: TStringStyle;
  borderTop: TStringStyle;
  borderLeft: TStringStyle;
  borderRight: TStringStyle;
  boxShadow: TStringStyle;
  fontFamily: TStringStyle;
  fontWeight: TStringStyle;
  animation: TStringStyle;
  backgroundImage: TStringStyle;
  backgroundSize: TStringStyle;
  backgroundPosition: TStringStyle;
  backdropFilter: TStringStyle;
  filter: TStringStyle;
  transform: TStringStyle;
  transition: TStringStyle;
  outline: TStringStyle;
  clip: TStringStyle;
  flexFlow: TStringStyle;
  textShadow: TStringStyle;
  content: TStringStyle;

  // color
  color: TColorStyle;
  backgroundColor: TColorStyle;
  borderColor: TColorStyle;
}

interface IStyleArgsAtoms extends IStyleArgsAtomsBase, IAtomsTool {}

interface IStyleAtoms extends IStyleArgsAtoms, INoneArgsAtoms, IComposeStyle, IStyleBuilder{
  _result: IJson<string | (()=>string)>;
  _styles: IStyleComponent[];
}
interface IPseudoBuilder extends IBuilderParameter {
  type: 'pseudo',
  exe (dom: HTMLElement): void;
}

export { IAtomsTool, IBuilderParameter, IComposeStyle, IComputedItem, IJson, INoneArgsAtoms, IPseudoBuilder, IReactBase, IReactBinding, IReactBindingTemplate, IReactBindingTemplateFactory, IReactBuilder, IReactContext, IReactItem, IReactObject, IReactWrap, IStyleArgsAtoms, IStyleArgsAtomsBase, IStyleAtoms, IStyleBuilder, IStyleComponent, IStyleComponentArray, TComputedFunc, TI, TOnChangeFunc, TPosition, TReactContextType, TReactionItem, TReactionValue, TStyleJsonValue, TStyleValue, TTextDeco, TUnit, delay, forceUpdate, getFuncArgs, getListeners, index, isStringTemplateArray, join, json, reactValue, replaceValue, splitTwoPart, subscribe, value, version };
